/*
 * XWIN.C
 *
 * (c)Copyright 2002, Matthew Dillon, see COPYRIGHT file.
 * $Backplane: dtv/xwin.c,v 1.15 2002/02/11 03:13:57 dillon Exp $
 */

#include "defs.h"
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xmd.h>
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86dgastr.h>

Prototype void FrameInit(TVCap *tv, XInfo *x, int shift, int flags);
Prototype void FrameOpen(TVCap *tv, XInfo *x);
Prototype void FrameDisplay(XInfo *x);
Prototype void FrameConvert(TVCap *tv, XInfo *x);
Prototype void FrameUserControl(TVCap *tv, XInfo *x, int fd);

static void FrameConvert3216(TVCap *tv, XInfo *x);
static void FrameConvertDirect(TVCap *tv, XInfo *x);
static void FrameConvert1632(TVCap *tv, XInfo *x);
static int xDgaQueryExtension(Display *d);

void
FrameInit(TVCap *tv, XInfo *x, int shift, int flags)
{
    bzero(x, sizeof(XInfo));

    x->x_Display = XOpenDisplay(NULL);
    if (x->x_Display == NULL)
	fatal("Unable to open X window");
    if (shift < 0) {
	x->Width = tv->tv_XSize << -shift;
	x->Height = tv->tv_YSize << -shift;
    } else {
	x->Width = tv->tv_XSize >> shift;
	x->Height = tv->tv_YSize >> shift;
    }
    x->Shift = shift;	/* may be negative */
    x->Flags = flags;
}

void
FrameOpen(TVCap *tv, XInfo *x)
{
    Display *d = x->x_Display;
    Window w;
    XImage *image;
    XVisualInfo *vi;
    XVisualInfo vpref;
    int i;
    int n;
    int use_shm;
    int use_dga;
    int displayIsLocal;
    char *dga_base = NULL;
    int dga_pitch = 0;
    int dga_bank = 0;
    int dga_ram = 0;

    w = XCreateSimpleWindow(
	    d,
	    DefaultRootWindow(d),
	    0,
	    0,
	    x->Width,
	    x->Height + 40,
	    1,
	    -1,
	    0
    );
    XClearWindow(d, w);
    XMapWindow(d, w);
    x->x_MWindow = w;

    w = XCreateSimpleWindow(
	    d,
	    x->x_MWindow,
	    0,
	    0,
	    x->Width,
	    x->Height,
	    0,
	    0,
	    0
    );
    XClearWindow(d, w);
    XMapWindow(d, w);
    x->x_VWindow = w;

    w = XCreateSimpleWindow(
	    d,
	    x->x_MWindow,
	    0,
	    x->Height,
	    x->Width,
	    40,
	    0,
	    0,
	    0
    );
    XClearWindow(d, w);
    XMapWindow(d, w);
    x->x_CWindow = w;

    XFlush(d);
    x->x_GC = XCreateGC(x->x_Display, x->x_VWindow, 0, NULL);

    vpref.screen = DefaultScreen(d);
    vi = XGetVisualInfo(x->x_Display, VisualScreenMask, &vpref, &n);

    for (i = 0; i < n; ++i) {
	fprintf(stderr, "visual %p depth %d/%d\n", vi, vi[i].depth, vi[i].bits_per_rgb);
    }

    /*
     * This is really stupid.  X should have a call that just tells me
     * whether the display is *truely* local or not
     */
    {
	const char *dname = getenv("DISPLAY");
	if (dname[0] != ':')
	    displayIsLocal = 0;
	else
	    displayIsLocal = 1;
    }

    /*
     * Can we use DGA?	XXX WORK IN PROGRESS
     *
     * Note: for a client/server setup DGA may not actually work better
     * because decompression does a read-modify-write and this can be
     * extremely slow compared to a direct bitmap copy (write only to the
     * frame buffer).
     */
    use_dga = xDgaQueryExtension(x->x_Display);
    use_dga = False;	/* XXX */
    if (use_dga == True && !displayIsLocal)
	use_dga = False;
    if (use_dga == True) {
	int flags = 0;

	XF86DGAQueryDirectVideo(x->x_Display, vi[0].screen, &flags);
	printf("FLAGS %08x\n", flags);
	if ((flags & XF86DGADirectPresent) == 0)
	    use_dga = False;
    }
    /*
     * WORK IN PROGRESS.
     */
    if (use_dga == True) {
	XF86DGAGetVideo(
	    x->x_Display,
	    vi[0].screen,
	    &dga_base,
	    &dga_pitch,
	    &dga_bank,
	    &dga_ram
	);
	dga_ram *= 1024;
	if (dga_base == NULL)
	    use_dga = False;
    }
    if (use_dga)
	use_shm = False;
    else
	use_shm = XShmQueryExtension(x->x_Display);
    if (use_shm == True && !displayIsLocal)
	use_shm = False;
    if (use_shm == True) {
	image = XShmCreateImage(
		    x->x_Display,
		    vi[0].visual,
		    vi[0].depth,
		    ZPixmap,
		    NULL,
		    &x->shm_info,
		    x->Width, x->Height
		);
	x->shm_info.shmid = shmget(
		    IPC_PRIVATE,
		    image->bytes_per_line * image->height,
		    IPC_CREAT | 0777
		);
	x->shm_info.shmaddr = shmat(x->shm_info.shmid, 0, 0);
	x->shm_info.readOnly = True;
	if (!XShmAttach(x->x_Display, &x->shm_info)) {
	    fprintf(stderr, "Unable to attach shared memory\n");
	    exit(1);
	}
	bzero(x->shm_info.shmaddr, image->bytes_per_line * image->height);
	XFlush(x->x_Display);
	XSync(x->x_Display, False);
	shmctl(x->shm_info.shmid, IPC_RMID, NULL);
	image->data = x->shm_info.shmaddr;
	x->using_shm = 1;
	x->BytesPerPix = image->bytes_per_line / x->Width;
    } else {
	image = XCreateImage(
		    x->x_Display,
		    vi[0].visual,
		    vi[0].depth,
		    ZPixmap,
		    0,
		    NULL,
		    x->Width,
		    x->Height,
		    BitmapPad(x->x_Display),
		    0
		);
	x->BytesPerPix = image->bytes_per_line / x->Width;
	if (use_dga == True) {
	    int row;

	    image->data = dga_base;
	    image->bytes_per_line = dga_pitch * x->BytesPerPix;
	    x->Flags |= XF_DGADIRECT;
	    fprintf(stderr, "using DGADIRECT\n");
	    for (row = 0; row < image->height; ++row) {
		bzero(
		    image->data + row * image->bytes_per_line, 
		    x->BytesPerPix * x->Width
		);
	    }
	} else {
	    image->data = malloc(image->bytes_per_line * image->height);
	    bzero(image->data, image->bytes_per_line * image->height);
	}
    }
    x->x_Image = image;
    if (x->Width >= tv->tv_XSize &&
	x->Height >= tv->tv_YSize &&
	x->BytesPerPix == tv->tv_PixSize &&
	x->Shift == 0 &&
	(x->Flags & XF_NODIRECT) == 0
    ) {
	printf("USING IMGDIRECT\n");
	x->Flags |= XF_DIRECT;
    }
}

void
FrameDisplay(XInfo *x)
{
    if (x->Flags & XF_DGADIRECT)
	return;
    if (x->shm_info.shmaddr) {
	XShmPutImage(
	    x->x_Display, x->x_VWindow, x->x_GC, x->x_Image, 
	    0, 0,
	    0, 0,
	    x->Width, x->Height,
	    False
	);
    } else {
	XPutImage(
	    x->x_Display, x->x_VWindow, x->x_GC, x->x_Image, 
	    0, 0,
	    0, 0,
	    x->Width, x->Height
	);
    }
    XFlush(x->x_Display);
}

/*
 * FrameUserControl() - setup our X controls
 *
 *	Controls will transmit command codes over the specified descriptor,
 *	typically a pipe or the direct socket to the server.
 */
void
FrameUserControl(TVCap *tv, XInfo *x, int fd)
{
    pid_t pid;
    int i;
    char *idStr = NULL;
    char *chStr = NULL;

    /*
     * Parent or error
     */
    if ((pid = fork()) != 0)
	return;

    /*
     * Child
     */
    dup2(fd, 1);
    for (i = 3; i < 64; ++i)
	close(i);
    asprintf(&idStr, "%ld", x->x_CWindow);
    asprintf(&chStr, "%d", (int)tv->tv_Channel);
    execlp("wish8.4", "wish", "/usr/local/etc/dtv-xevent.tk", "-use", idStr, "--", chStr, NULL);
    execlp("wish", "wish", "/usr/local/etc/dtv-xevent.tk", "-use", idStr, "--", chStr, NULL);
    _exit(0);
}

/*
 * FrameConvert() - convert TV frame into XImage frame
 */
void
FrameConvert(TVCap *tv, XInfo *x)
{
    /*
     * If XF_DIRECT is set we decompressed directly into the XImage, so
     * we do not have to do anything here.
     */
    if (x->Flags & XF_DIRECT)
	return;

    /*
     * Match up pixel size
     */
    if (tv->tv_PixSize == 4) {
	switch (x->BytesPerPix) {
	case 1:
	    assert(0);
	    break;
	case 2:
	    FrameConvert3216(tv, x);
	    break;
	case 3:
	    assert(0);
	    break;
	case 4:
	    FrameConvertDirect(tv, x);
	    break;
	}
    } else if (tv->tv_PixSize == 2) {
	switch (x->BytesPerPix) {
	case 1:
	    assert(0);
	    break;
	case 2:
	    FrameConvertDirect(tv, x);
	    break;
	case 3:
	    assert(0);
	    break;
	case 4:
	    FrameConvert1632(tv, x);
	    break;
	}
    }
}

static void
FrameConvert3216(TVCap *tv, XInfo *x)
{
    int rows;
    int mult = (x->Shift < 0) ? -(1 << -x->Shift) : (1 << x->Shift);
    const int rmask = 0x1F;
    const int gmask = 0x3F;
    const int bmask = 0x1F;
    const int rshift = 11;
    const int gshift = 5;
    const int bshift = 0;
    unsigned char *sptr = (tv->tv_Source == SOURCE_CAPBUF) ? tv->tv_CapBuf : tv->tv_CmpBuf;
    unsigned char *dbase = x->x_Image->data;

    rows = 0;
    while (rows < tv->tv_YSize) {
	unsigned char *dptr = dbase;
	int i;

	if (mult == 1) {
	    for (i = tv->tv_XSize - 1; i >= 0; --i) {
		*(int16_t *)dptr = (((sptr[0] >> 3) & bmask) << bshift) |
				   (((sptr[1] >> 2) & gmask) << gshift) |
				   (((sptr[2] >> 3) & rmask) << rshift);
		sptr += 4;
		dptr += 2;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    ++rows;
	} else if (mult < 0) {
	    for (i = tv->tv_XSize - 1; i >= 0; --i) {
		int j;

		if (x->Flags & XF_SKIPLINES) {
		    *(int16_t *)dptr = (((sptr[0] >> 3) & bmask) << bshift) |
				       (((sptr[1] >> 2) & gmask) << gshift) |
				       (((sptr[2] >> 3) & rmask) << rshift);
		    sptr += 4;
		    dptr += 2 * -mult;
		    continue;
		}
		for (j = mult; j < 0; ++j) {
		    *(int16_t *)dptr = (((sptr[0] >> 3) & bmask) << bshift) |
				       (((sptr[1] >> 2) & gmask) << gshift) |
				       (((sptr[2] >> 3) & rmask) << rshift);
		    dptr += 2;
		}
		sptr += 4;
	    }
	    for (i = mult + 1; i < 0; ++i) {
		if ((x->Flags & XF_SKIPLINES) == 0)
		    bcopy(dbase, dbase + x->x_Image->bytes_per_line, x->BytesPerPix * x->Width);
		dbase += x->x_Image->bytes_per_line;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    ++rows;
	} else {
	    for (i = 0; i < tv->tv_XSize; ) {
		int r = 0;
		int g = 0;
		int b = 0;
		do {
		    b += sptr[0];
		    g += sptr[1];
		    r += sptr[2];
		    sptr += 4;
		    ++i;
		} while (i % mult != 0);
		*(int16_t *)dptr =
		    (((b / mult / 8) & bmask) << bshift) |
		    (((g / mult / 4) & gmask) << gshift) |
		    (((r / mult / 8) & rmask) << rshift);
		dptr += 2;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    sptr += tv->tv_XSize * (mult - 1) * 4;
	    rows += mult;
	}
    }
}


static void
FrameConvertDirect(TVCap *tv, XInfo *x)
{
    int rows;
    int mult = (x->Shift < 0) ? -(1 << -x->Shift) : (1 << x->Shift);
    unsigned char *sptr = (tv->tv_Source == SOURCE_CAPBUF) ? tv->tv_CapBuf : tv->tv_CmpBuf;
    unsigned char *dbase = x->x_Image->data;

    rows = 0;
    while (rows < tv->tv_YSize) {
	unsigned char *dptr = dbase;
	int i;

	if (mult == 1) {
	    bcopy(sptr, dptr, x->BytesPerPix * x->Width);
	    dbase += x->x_Image->bytes_per_line;
	    sptr += tv->tv_XSize * tv->tv_PixSize;
	    ++rows;
	} else if (mult < 0) {
	    switch(tv->tv_PixSize) {
	    case 2:
		for (i = tv->tv_XSize - 1; i >= 0; --i) {
		    u_int16_t s = *(u_int16_t *)sptr;
		    int j;
		    for (j = mult; j < 0; ++j) {
			*(u_int16_t *)dptr = s;
			dptr += 2;
		    }
		    sptr += 2;
		}
		break;
	    case 4:
		for (i = tv->tv_XSize - 1; i >= 0; --i) {
		    u_int32_t s = *(u_int32_t *)sptr;
		    int j;

		    if (x->Flags & XF_SKIPLINES) {
			*(u_int32_t *)dptr = s;
			dptr += 4 * -mult;
			sptr += 4;
			continue;
		    }
		    for (j = mult; j < 0; ++j) {
			*(u_int32_t *)dptr = s;
			dptr += 4;
		    }
		    sptr += 4;
		}
		break;
	    default:
		assert(0);
	    }
	    for (i = mult + 1; i < 0; ++i) {
		if ((x->Flags & XF_SKIPLINES) == 0)
		    bcopy(dbase, dbase + x->x_Image->bytes_per_line, x->BytesPerPix * x->Width);
		dbase += x->x_Image->bytes_per_line;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    ++rows;
	} else {
	    for (i = 0; i < tv->tv_XSize; ) {
		int r = 0;
		int g = 0;
		int b = 0;
		do {
		    b += sptr[0];
		    g += sptr[1];
		    r += sptr[2];
		    sptr += 4;
		    ++i;
		} while (i % mult != 0);
		*(int32_t *)dptr = ((u_int8_t)(r / mult) << 16) |
				    ((u_int8_t)(g / mult) << 8) | 
				    (u_int8_t)(b / mult);
		dptr += 4;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    sptr += tv->tv_XSize * (mult - 1) * tv->tv_PixSize;
	    rows += mult;
	}
    }
}

static void
FrameConvert1632(TVCap *tv, XInfo *x)
{
    int rows;
    int mult = (x->Shift < 0) ? -(1 << -x->Shift) : (1 << x->Shift);
    unsigned char *sptr = (tv->tv_Source == SOURCE_CAPBUF) ? tv->tv_CapBuf : tv->tv_CmpBuf;
    unsigned char *dbase = x->x_Image->data;

    rows = 0;
    while (rows < tv->tv_YSize) {
	unsigned char *dptr = dbase;
	int i;

	if (mult == 1) {
	    for (i = tv->tv_XSize - 1; i >= 0; --i) {
		u_int16_t s = *(u_int16_t *)sptr;
		*(int32_t *)dptr = ((s & 0x001F) << 3) |
				   ((s & 0x03E0) << 6) |
				   ((s & 0x7C00) << 9);
		sptr += 2;
		dptr += 4;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    ++rows;
	} else if (mult < 0) {
	    for (i = tv->tv_XSize - 1; i >= 0; --i) {
		u_int16_t s = *(u_int16_t *)sptr;
		int j;

		for (j = mult; j < 0; ++j) {
		    *(int32_t *)dptr = ((s & 0x001F) << 3) |
				       ((s & 0x03E0) << 6) |
				       ((s & 0x7C00) << 9);
		    dptr += 4;
		}
		sptr += 2;
	    }
	    for (i = -mult; i > 0; --i) {
		bcopy(dbase, dbase + x->x_Image->bytes_per_line, x->BytesPerPix * x->Width);
		dbase += x->x_Image->bytes_per_line;
	    }
	    ++rows;
	} else {
	    for (i = 0; i < tv->tv_XSize; ) {
		int r = 0;
		int g = 0;
		int b = 0;
		do {
		    u_int16_t s = *(u_int16_t *)sptr;
		    b += s & 0x001F;
		    g += s & 0x03E0;
		    r += s & 0x7C00;
		    sptr += 2;
		    ++i;
		} while (i % mult != 0);
		*(int32_t *)dptr = ((b << 3) / mult) |
				    ((g << 6) / mult) |
				    ((r << 9) / mult);
		dptr += 4;
	    }
	    dbase += x->x_Image->bytes_per_line;
	    sptr += tv->tv_XSize * (mult - 1) * tv->tv_PixSize;
	    rows += mult;
	}
    }
}

static int
xDgaQueryExtension(Display *d)
{
    int op = 0;
    int ev = 0;
    int err = 0;
    int use_dga;

    use_dga = XQueryExtension(d, XF86DGANAME, &op, &ev, &err);
    return(use_dga);
}

